/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation. You find a copy of this
 *  license in the file COPYRIGHT in the root directory of this release.
 *
 *  THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL,
 *  BUT WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
 *  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *
 *  Description:
 *
 *    This is an interface layer between C and Tcl/Tk.
 *
 *  Author: Tom Zoerner
 *
 *  $Id: gui_if.c,v 1.3 2005/05/31 20:14:32 tom Exp tom $
 */

#include <stdlib.h>
#include <stdio.h>
#include <locale.h>

#include <tcl.h>
#include <tk.h>

#include "gui.h"  // generated by tcl2c from gui.tcl

#include "sprio-test.h"
#include "syserrmsg.h"
#include "gui_if.h"

#define eval_check(INTERP,CMD) \
   do { \
      if (Tcl_EvalEx((INTERP), (CMD), -1, 0) != TCL_OK) \
         fprintf(stderr,"Tcl command: %s\nError: %s", (CMD), Tcl_GetStringResult(interp)); \
   } while(0)

// prior to 8.4 there's a SEGV when evaluating const scripts (Tcl tries to modify the string)
#if (TCL_MAJOR_VERSION > 8) || ((TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION >= 4))
# define TCL_EVAL_CONST(INTERP, SCRIPT) Tcl_EvalEx(INTERP, SCRIPT, -1, TCL_EVAL_GLOBAL)
#else
# define TCL_EVAL_CONST(INTERP, SCRIPT) Tcl_VarEval(INTERP, "uplevel #0 {", (char *) SCRIPT, "}", NULL)
#endif

// In Tcl 8.4 several function pointer args were made const.
// define a macro for backwards compatibility; this is included with 8.4's tcl.h
#ifndef CONST84
# if defined(TCL_MAJOR_VERSION) && defined(TCL_MINOR_VERSION)
#  if (TCL_MAJOR_VERSION < 8) || \
      ((TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION <= 3))
#   define CONST84
#  else
#   define CONST84  const
#  endif
# endif
#endif

#define TCL_COMM_BUF_SIZE 2000

static Tcl_Interp * interp;               // Tcl interpreter interface
static char comm[TCL_COMM_BUF_SIZE + 1];  // Command buffer (one extra byte for overflow detection)

// ---------------------------------------------------------------------------
// Helper function to print error message in GUI
//
static void DisplayErrMsg( const char * p_msg )
{
   sprintf(comm, ".all.cur.status.txt delete 1.0 end\n"
                 ".all.cur.status.txt insert end {%s}\n", p_msg);
   eval_check(interp, comm);
}

// ---------------------------------------------------------------------------
// Called by Tcl script to open or change the device
//
static int GuiIf_OpenDevice( ClientData ttp, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
   const char * const pUsage = "Usage: C_OpenDevice <path>";
   char * p_driver;
   char * p_card;
   char * p_err_msg;
   Tcl_Obj * pResultList;
   Tcl_DString dstr;
   int  result;

   if (objc != 1+1)
   {  // parameter count or format is invalid
      Tcl_SetResult(interp, (char *)pUsage, TCL_STATIC);
      result = TCL_ERROR;
   }
   else
   {
      p_err_msg = NULL;
      if ( PrioOpen(Tcl_GetString(objv[1]), &p_driver, &p_card, &p_err_msg) )
      {
         Tcl_DStringInit(&dstr);
         pResultList = Tcl_NewListObj(0, NULL);

         Tcl_ExternalToUtfDString(NULL, p_driver, -1, &dstr);
         Tcl_ListObjAppendElement(interp, pResultList, Tcl_NewStringObj(Tcl_DStringValue(&dstr), -1));

         Tcl_ExternalToUtfDString(NULL, p_card, -1, &dstr);
         Tcl_ListObjAppendElement(interp, pResultList, Tcl_NewStringObj(Tcl_DStringValue(&dstr), -1));

         Tcl_SetObjResult(interp, pResultList);
         Tcl_DStringFree(&dstr);
      }

      if (p_err_msg != NULL)
      {
         DisplayErrMsg(p_err_msg);
         SystemErrorMessage_Reset(&p_err_msg);
      }

      result = TCL_OK;
   }
   return result;
}

// ---------------------------------------------------------------------------
// Called by Tcl script to query the current channel
//
static int GuiIf_QueryChannel( ClientData ttp, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
   const char * const pUsage = "Usage: C_QueryChannel <app_idx>";
   char * p_err_msg;
   int  app_idx;
   int  freq;
   char *p_norm;
   Tcl_Obj * pResultList;
   int  result;

   if ( (objc != 1+1) ||
        (Tcl_GetIntFromObj(interp, objv[1], &app_idx) != TCL_OK) )
   {  // parameter count or format is invalid
      Tcl_SetResult(interp, (char *)pUsage, TCL_STATIC);
      result = TCL_ERROR;
   }
   else
   {
      p_err_msg = NULL;
      if ( PrioQueryChannel(app_idx, &freq, &p_norm, &p_err_msg) )
      {
         pResultList = Tcl_NewListObj(0, NULL);

         Tcl_ListObjAppendElement(interp, pResultList, Tcl_NewIntObj(freq));
         Tcl_ListObjAppendElement(interp, pResultList, Tcl_NewStringObj(p_norm, -1));

         Tcl_SetObjResult(interp, pResultList);
      }

      if (p_err_msg != NULL)
      {
         DisplayErrMsg(p_err_msg);
         SystemErrorMessage_Reset(&p_err_msg);
      }

      result = TCL_OK;
   }
   return result;
}

// ---------------------------------------------------------------------------
// Called by Tcl script to attempt channel change
//
static int GuiIf_ChangeChannel( ClientData ttp, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
   const char * const pUsage = "Usage: C_ChangeChannel <app_idx> <freq> <norm>";
   char * p_err_msg;
   int  app_idx;
   int  freq;
   int  result;

   if ( (objc != 1+3) ||
        (Tcl_GetIntFromObj(interp, objv[1], &app_idx) != TCL_OK) ||
        (Tcl_GetIntFromObj(interp, objv[2], &freq) != TCL_OK) )
   {  // parameter count or format is invalid
      Tcl_SetResult(interp, (char *)pUsage, TCL_STATIC);
      result = TCL_ERROR;
   }
   else
   {
      p_err_msg = NULL;
      if ( PrioChangeChannel(app_idx, freq, Tcl_GetString(objv[3]), &p_err_msg) )
      {
         DisplayErrMsg("OK");
      }

      if (p_err_msg != NULL)
      {
         DisplayErrMsg(p_err_msg);
         SystemErrorMessage_Reset(&p_err_msg);
      }

      result = TCL_OK;
   }
   return result;
}

// ---------------------------------------------------------------------------
// Called by Tcl script to query the priority
//
static int GuiIf_QueryPriority( ClientData ttp, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
   const char * const pUsage = "Usage: C_QueryPriority";
   char * p_err_msg;
   int  app_idx;
   int  prio;
   int  result;

   if ( (objc != 1+1) ||
        (Tcl_GetIntFromObj(interp, objv[1], &app_idx) != TCL_OK) )
   {  // parameter count or format is invalid
      Tcl_SetResult(interp, (char *)pUsage, TCL_STATIC);
      result = TCL_ERROR;
   }
   else
   {
      p_err_msg = NULL;
      if ( PrioQuery(app_idx, &prio, &p_err_msg) )
      {
         Tcl_SetObjResult(interp, Tcl_NewIntObj(prio));
      }
      else
         Tcl_SetObjResult(interp, Tcl_NewIntObj(0));

      if (p_err_msg != NULL)
      {
         DisplayErrMsg(p_err_msg);
         SystemErrorMessage_Reset(&p_err_msg);
      }

      result = TCL_OK;
   }
   return result;
}

// ---------------------------------------------------------------------------
// Called by Tcl script to set the priority
//
static int GuiIf_SetPriority( ClientData ttp, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
   const char * const pUsage = "Usage: C_SetPriority <app_idx> <prio>";
   char * p_err_msg;
   int  app_idx;
   int  prio;
   int  result;

   if ( (objc != 1+2) ||
        (Tcl_GetIntFromObj(interp, objv[1], &app_idx) != TCL_OK) ||
        (Tcl_GetIntFromObj(interp, objv[2], &prio) != TCL_OK) )
   {  // parameter count or format is invalid
      Tcl_SetResult(interp, (char *)pUsage, TCL_STATIC);
      result = TCL_ERROR;
   }
   else
   {
      p_err_msg = NULL;
      if ( PrioSet(app_idx, prio, &p_err_msg) )
      {
         // ok - nothing to do
      }

      if (p_err_msg != NULL)
      {
         DisplayErrMsg(p_err_msg);
         SystemErrorMessage_Reset(&p_err_msg);
      }

      result = TCL_OK;
   }
   return result;
}

// ---------------------------------------------------------------------------
// Initialize the Tcl/Tk interpreter
//
static int ui_init( int argc, char **argv )
{
   char *args;

   if (argc >= 1)
   {
      Tcl_FindExecutable(argv[0]);
   }

   interp = Tcl_CreateInterp();

   if (argc > 1)
   {
      args = Tcl_Merge(argc - 1, (CONST84 char **) argv + 1);
      Tcl_SetVar(interp, "argv", args, TCL_GLOBAL_ONLY);
      Tcl_SetVar2Ex(interp, "argc", NULL, Tcl_NewIntObj(argc - 1), TCL_GLOBAL_ONLY);
   }
   Tcl_SetVar(interp, "argv0", argv[0], TCL_GLOBAL_ONLY);

   Tcl_SetVar(interp, "tcl_library", TCL_LIBRARY_PATH, TCL_GLOBAL_ONLY);
   Tcl_SetVar(interp, "tk_library", TK_LIBRARY_PATH, TCL_GLOBAL_ONLY);
   Tcl_SetVar(interp, "tcl_interactive", "0", TCL_GLOBAL_ONLY);

   Tcl_Init(interp);
   if (Tk_Init(interp) != TCL_OK)
   {
      fprintf(stderr, "Failed to initialise the Tk library at '%s' - exiting.\nTk error message: %s\n",
                      TK_LIBRARY_PATH, Tcl_GetStringResult(interp));
      exit(1);
   }

   if (TCL_EVAL_CONST(interp, gui_tcl_static) != TCL_OK)
   {
      fprintf(stderr, "LoadTcl-Init: failed to compile gui.tcl: %s", Tcl_GetStringResult(interp));
   }

   sprintf(comm, "wm title . {video4linux device priority check}\n"
                 "wm resizable . 0 1\n"
                 "wm iconname . {v4l-prio-check}\n");
   eval_check(interp, comm);

   Tcl_ResetResult(interp);
   return (TRUE);
}

// ---------------------------------------------------------------------------
// GUI main loop
//
void GuiMainLoop( int argc, char *argv[] )
{
   // mark Tcl/Tk interpreter as uninitialized
   interp = NULL;

   // set up the user-configured locale
   setlocale(LC_ALL, "");
   setlocale(LC_NUMERIC, "C");  // required for Tcl or parsing of floating point numbers fails

   // initialize Tcl interpreter and compile all scripts
   // Tk is only initialized if a GUI will be opened
   ui_init(argc, argv);

   Tcl_CreateObjCommand(interp, "C_OpenDevice", GuiIf_OpenDevice, (ClientData) NULL, NULL);
   Tcl_CreateObjCommand(interp, "C_QueryPriority", GuiIf_QueryPriority, (ClientData) NULL, NULL);
   Tcl_CreateObjCommand(interp, "C_SetPriority", GuiIf_SetPriority, (ClientData) NULL, NULL);
   Tcl_CreateObjCommand(interp, "C_QueryChannel", GuiIf_QueryChannel, (ClientData) NULL, NULL);
   Tcl_CreateObjCommand(interp, "C_ChangeChannel", GuiIf_ChangeChannel, (ClientData) NULL, NULL);

   eval_check(interp, "CreateMainWindow\n");

   // wait until window is open and everything displayed
   while ( (Tk_GetNumMainWindows() > 0) &&
           Tcl_DoOneEvent(TCL_ALL_EVENTS | TCL_DONT_WAIT) )
      ;

   if (Tk_GetNumMainWindows() > 0)
   {
      while (Tk_GetNumMainWindows() > 0)
      {
         if (Tcl_DoOneEvent(TCL_DONT_WAIT) == 0)
         {
            // schedule idle events
         }
      }

      if (Tk_GetNumMainWindows() > 0)
      {
         // Tk shutdown
      }
   }
   else
      fprintf(stderr, "could not open the main window - exiting.\n");
}

