/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation. You find a copy of this
 *  license in the file COPYRIGHT in the root directory of this release.
 *
 *  THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL,
 *  BUT WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
 *  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *
 *  Description:
 *
 *    This is a little application which allows to check if ioctl
 *    VIDIOC_S_PRIORITY works with your v4l(2) driver.
 *
 *  Author: Tom Zoerner
 *
 *  $Id: sprio-test.c,v 1.3 2005/05/31 20:14:11 tom Exp tom $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>

#include "syserrmsg.h"
#include "gui_if.h"
#include "sprio-test.h"

#ifndef PATH_VIDEODEV_H
# include "videodev.h"
#else
# include PATH_VIDEODEV_H
#endif

/* same as ioctl(), but repeat if interrupted */
#define IOCTL(fd, cmd, data)                                            \
({ int __result; do __result = ioctl(fd, cmd, data);                    \
   while ((__result == -1L) && (errno == EINTR)); __result; })

typedef struct
{
   int          fd_1;
   int          fd_2;
} SPRIO_STATE;

static SPRIO_STATE sps;

// ---------------------------------------------------------------------------
// Query the current channel
// - note the app idx is not really useful in this case since both will return
//   the same value
//
sbool PrioQueryChannel( int app_idx, int * p_freq, char ** pp_norm, char ** pp_err_msg )
{
   struct video_channel vchan;
   long lfreq;
   int fd;
   sbool result = FALSE;

   fd = ((app_idx == 0) ? sps.fd_1 : sps.fd_2);

   if (fd != -1)
   {
      lfreq = 0;
      if (IOCTL(fd, VIDIOCGFREQ, &lfreq) == 0)
      {
         memset(&vchan, 0, sizeof(vchan));

         if (IOCTL(fd, VIDIOCGCHAN, &vchan) == 0)
         {
            if (vchan.norm == VIDEO_MODE_NTSC)
               *pp_norm = "NTSC";
            else if (vchan.norm == VIDEO_MODE_SECAM)
               *pp_norm = "SECAM";
            else //if (vchan.norm == VIDEO_MODE_PAL)
               *pp_norm = "PAL";

            if ( (vchan.type & VIDEO_TYPE_TV) &&
                 (vchan.flags & VIDEO_VC_TUNER) )
            {
               *p_freq = lfreq;
            }
            else
               *p_freq = 0;

            result = TRUE;
         }
         else
            SystemErrorMessage_Set(pp_err_msg, errno, "Failed to query input channel (v4l ioctl VIDIOCGCHAN): ", NULL);
      }
      else
         SystemErrorMessage_Set(pp_err_msg, errno, "Failed to query input channel (v4l ioctl VIDIOCGFREQ): ", NULL);
   }
   return result;
}

// ---------------------------------------------------------------------------
// Attempt to switch the TV channel
//
sbool PrioChangeChannel( int app_idx, int freq, const char * p_norm, char ** pp_err_msg )
{
   struct video_channel vchan;
   long lfreq;
   int norm;
   int fd;
   sbool result = FALSE;

   fd = ((app_idx == 0) ? sps.fd_1 : sps.fd_2);

   if (fd != -1)
   {
      if (strcasecmp(p_norm, "NTSC") == 0)
         norm = VIDEO_MODE_NTSC;
      else if (strcasecmp(p_norm, "SECAM") == 0)
         norm = VIDEO_MODE_SECAM;
      else //if (strcasecmp(p_norm, "PAL") == 0)
         norm = VIDEO_MODE_PAL;

      // get current config of the selected chanel
      memset(&vchan, 0, sizeof(vchan));
      vchan.channel = 0;
      if (IOCTL(fd, VIDIOCGCHAN, &vchan) == 0)
      {  // channel index is valid

         vchan.norm    = norm;
         vchan.channel = 0;

         if (IOCTL(fd, VIDIOCSCHAN, &vchan) == 0)
         {
            if ( (vchan.type & VIDEO_TYPE_TV) && (vchan.flags & VIDEO_VC_TUNER) )
            {
               // Set the tuner frequency
               lfreq = freq;
               if (IOCTL(fd, VIDIOCSFREQ, &lfreq) == 0)
               {
                  result = TRUE;
               }
               else
                  SystemErrorMessage_Set(pp_err_msg, errno, "Failed to set tuner frequency (v4l ioctl VIDIOCSFREQ): ", NULL);
            }
            else
               SystemErrorMessage_Set(pp_err_msg, 0, "Seems OK, but warning: channel #0 is not a TV tuner", NULL);
         }
         else
            SystemErrorMessage_Set(pp_err_msg, errno, "Failed to switch TV input channel: ", NULL);
      }
      else
         SystemErrorMessage_Set(pp_err_msg, errno, "Failed to query input channel: ", NULL);
   }
   return result;
}

// ---------------------------------------------------------------------------
// Change the priority
//
sbool PrioSet( int app_idx, int prio, char ** pp_err_msg )
{
   int fd;
   sbool result = FALSE;

   fd = ((app_idx == 0) ? sps.fd_1 : sps.fd_2);

   if (fd != -1)
   {
#ifdef VIDIOC_S_PRIORITY
      enum v4l2_priority v4l2_prio = prio;

      if (IOCTL(fd, VIDIOC_S_PRIORITY, &v4l2_prio) == 0)
      {
         // ok
      }
      else
      {
         if (errno == EINVAL)
            SystemErrorMessage_Set(pp_err_msg, 0,
                                   "Priorities are not supported by this device", NULL);
         else
            SystemErrorMessage_Set(pp_err_msg, errno,
                                   "Failed to change priority threshold: ", NULL);
      }
#else  // no VIDIOC_S_PRIORITY
#endif
   }
   return result;
}

// ---------------------------------------------------------------------------
// Query the current priority threshold
//
sbool PrioQuery( int app_idx, int * p_prio, char ** pp_err_msg )
{
   int fd;
   sbool result = FALSE;

   fd = ((app_idx == 0) ? sps.fd_1 : sps.fd_2);

   if (fd != -1)
   {
#ifdef VIDIOC_S_PRIORITY
      enum v4l2_priority prio = -1;

      if (IOCTL(fd, VIDIOC_G_PRIORITY, &prio) == 0)
      {
         *p_prio = prio;
         result = TRUE;
      }
      else
      {
         if (errno == EINVAL)
            SystemErrorMessage_Set(pp_err_msg, 0,
                                   "Priorities are not supported by this device", NULL);
         else
            SystemErrorMessage_Set(pp_err_msg, errno,
                                   "Failed to query current priority threshold: ", NULL);
      }
#else
      SystemErrorMessage_Set(pp_err_msg, 0,
                             "Compiled without support for VIDIOC_S_PRIORITY: check your videodev2.h header file", NULL);
#endif
   }
   return result;
}

// ---------------------------------------------------------------------------
// Close the device
//
void PrioClose( void )
{
   if (sps.fd_1 != -1)
      close(sps.fd_1);
   if (sps.fd_2 != -1)
      close(sps.fd_2);

   sps.fd_1 = sps.fd_2 = -1;
}

// ---------------------------------------------------------------------------
// Open the device, check if it's a video4linux device and return card info
//
sbool PrioOpen( const char * p_dev_path, char ** pp_driver, char ** pp_card, char ** pp_err_msg )
{
   static struct v4l2_capability  vcap2;
   struct video_capability vcap1;
   int  fd;
   sbool result = FALSE;

   PrioClose();

   fd = open(p_dev_path, O_RDWR);
   if (fd != -1)
   {
      // check if it's a v4l2 device at all
      memset(&vcap1, 0, sizeof(vcap1));
      memset(&vcap2, 0, sizeof(vcap2));

      if (IOCTL(fd, VIDIOC_QUERYCAP, &vcap2) != -1)
      {
         *pp_driver = vcap2.driver;
         *pp_card = vcap2.card;

         sps.fd_1 = fd;
         // open the device a 2nd time
         sps.fd_2 = open(p_dev_path, O_RDWR);
         if (sps.fd_2 != -1)
         {
            result = TRUE;
         }
         else
         {
            SystemErrorMessage_Set(pp_err_msg, errno,
                                   "Failed to open the device a 2nd time "
                                   "-> this is not a standards compliant v4l2 device!? Syscall error: ", NULL);
            close(fd);
            sps.fd_1 = -1;
         }
      }
      else if (IOCTL(fd, VIDIOCGCAP, &vcap1) != -1)
      {
         SystemErrorMessage_Set(pp_err_msg, 0,
                                "This driver appears to support only the 1st generation of the v4l API. "
                                "No priority handling is required here, since the device cannot be "
                                "opened multiple times.", NULL);
         close(fd);
      }
      else
      {
         SystemErrorMessage_Set(pp_err_msg, errno,
                                "Failed to query video device capabilities "
                                "-> this is probably not a video device; Syscall error: ", NULL);
         close(fd);
      }
   }
   else
   {
      SystemErrorMessage_Set(pp_err_msg, errno,
                             "Failed to open device: ", p_dev_path, NULL);
   }
   return result;
}

// ---------------------------------------------------------------------------
// entry point
//
int main( int argc, char *argv[] )
{
   // init the internal module state
   memset(&sps, 0, sizeof(sps));
   sps.fd_1 = sps.fd_2 = -1;

   GuiMainLoop(argc, argv);

   return 0;
}


